/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <string>
#include <sstream>
#include <cstdlib>

#include "composite-canvas-xrender.h"
#include "composite-window-xrender.h"
#include "options.h"
#include "log.h"

CompositeCanvasXRender::~CompositeCanvasXRender()
{
    release_resources();
}

bool
CompositeCanvasXRender::have_xrender()
{
    int ignore, major, minor;
    bool have_xrender = false;
        
    /* Check for the XShm extension */
    if (XQueryExtension(xdpy_, "RENDER", &ignore, &ignore, &ignore)) {
        if (XRenderQueryVersion(xdpy_, &major, &minor)) {
            Log::debug("XRender extension version %d.%d\n",
                       major, minor);
            have_xrender = true;
        }
    }

    return have_xrender;
}

bool
CompositeCanvasXRender::have_xdbe()
{
    int major, minor;
    bool have_xdbe = false;
        
    /* Check for the Xdbe extension */
    if (XdbeQueryExtension(xdpy_, &major, &minor)) {
        Log::debug("Xdbe extension version %d.%d\n", major, minor);
        have_xdbe = true;
    }
    else {
        Log::info("** Xdbe extension not supported. You may experience flickering.\n");
    }

    return have_xdbe;
}

bool
CompositeCanvasXRender::ensure_xdbe_back_buffer()
{
    XdbeScreenVisualInfo *xdbe_screen_vis_info;
    XWindowAttributes canvas_attr;
    int num_screens = 1;
    bool dbe_visual_found = false;

    if (xdbe_back_buffer_)
        return true;

    if (!have_xdbe())
        return false;

    /* Check if the canvas visual supports double-buffering */
    if (!XGetWindowAttributes(xdpy_, canvas_, &canvas_attr)) {
        Log::error("XGetWindowAttributes failed!\n");
        return false;
    }

    xdbe_screen_vis_info = XdbeGetVisualInfo(xdpy_, &canvas_, &num_screens);
    if (!xdbe_screen_vis_info || num_screens == 0) {
        Log::error("XdbeGetVisualInfo failed!\n");
        return false;
    }

    for (int i = 0; i < xdbe_screen_vis_info[0].count; i++) {
        XdbeVisualInfo *vis_info = &xdbe_screen_vis_info[0].visinfo[i];
        if (vis_info->visual == XVisualIDFromVisual(canvas_attr.visual)) {
            dbe_visual_found = true;
            break;
        }
    }

    XdbeFreeVisualInfo(xdbe_screen_vis_info);
 
    if (!dbe_visual_found) {
        Log::info("** The window visual does not support double buffering."
                  " You may experience flickering!\n");
        return false;
    }

    /* Allocate the back buffer */
    xdbe_back_buffer_ = XdbeAllocateBackBufferName(xdpy_, canvas_,
                                                   XdbeBackground);
    if (!xdbe_back_buffer_) {
        Log::error("XdbeAllocateBufferName failed!\n");
        return false;
    }

    return true;
}

bool
CompositeCanvasXRender::ensure_xrender_picture()
{
    Drawable drawable;

    if (xrender_picture_)
        return true;

    if (!have_xrender())
        return false;

    if (!ensure_xdbe_back_buffer())
        drawable = canvas_;
    else
        drawable = xdbe_back_buffer_;

    XWindowAttributes canvas_attr;

    if (!XGetWindowAttributes(xdpy_, canvas_, &canvas_attr)) {
        Log::error("XGetWindowAttributes failed!\n");
        return false;
    }

    /* TODO: should we free this when we are done? */
    XRenderPictFormat *pict_format;
        
    pict_format = XRenderFindVisualFormat(xdpy_, canvas_attr.visual);
    if (!pict_format) {
        Log::error("XRenderFindVisualFormat failed!\n");
        release_resources();
        return false;
    }

    xrender_picture_ = XRenderCreatePicture(xdpy_, drawable, pict_format,
                                            0, NULL);
    if (!xrender_picture_) {
        Log::error("XRenderCreatePicture failed!\n");
        release_resources();
        return false;
    }

    return true;
}

XVisualInfo *
CompositeCanvasXRender::get_canvas_xvisualinfo()
{
    int num_visuals;
    XVisualInfo vis_tmpl;
    XVisualInfo *vis_info;

    vis_tmpl.depth = 32;
    vis_tmpl.bits_per_rgb = 8;
    vis_info = XGetVisualInfo(xdpy_, VisualDepthMask | VisualBitsPerRGBMask,
                              &vis_tmpl, &num_visuals);
    if (!vis_info) {
        Log::error("Error: couldn't get X visual\n");
        return 0;
    }

    return vis_info;
}

void
CompositeCanvasXRender::reshape(int width, int height)
{
    CompositeCanvas::reshape(width, height);

    release_resources();
    ensure_xrender_picture();
}

bool
CompositeCanvasXRender::make_current()
{
    return ensure_xrender_picture();
}

void
CompositeCanvasXRender::release_resources()
{
    if (xrender_picture_)
        XRenderFreePicture(xdpy_, xrender_picture_);

    if (xdbe_back_buffer_)
        XdbeDeallocateBackBufferName(xdpy_, xdbe_back_buffer_);

    xrender_picture_ = 0;
    xdbe_back_buffer_ = 0;
}

CompositeWindow *
CompositeCanvasXRender::create_window(Window win)
{
    return new CompositeWindowXRender(xdpy_, win);
}

void
CompositeCanvasXRender::update_screen()
{
    if (xdbe_back_buffer_) {
        XdbeSwapInfo swap_info = {canvas_, XdbeBackground};
        XdbeSwapBuffers(xdpy_, &swap_info, 1);
    }
}

std::string
CompositeCanvasXRender::info_string()
{
    std::stringstream ss;
    int major, minor;

    ss << "    XRender ";
    if (XRenderQueryVersion(xdpy_, &major, &minor))
        ss << major << "." << minor << std::endl;
    else
        ss << "(Not Available)" << std::endl;

    return ss.str();
}
