/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <X11/extensions/Xrender.h>
#include <cmath>

#include "composite-test.h"
#include "composite-canvas-xrender.h"

CompositeTestXRender::CompositeTestXRender() :
    CompositeTest("xrender", "xrender")
{
    options_["filter"] = CompositeTest::Option("filter", "bilinear",
                                               "Filter to use [nearest, bilinear]");
}

void
CompositeTestXRender::init()
{
    XRenderColor half_transparent = {0x7fff, 0x7fff, 0x7fff, 0x7fff};
    half_transparent_mask_ = XRenderCreateSolidFill(canvas_->xdisplay(),
                                                    &half_transparent);
}

void
CompositeTestXRender::deinit()
{
    if (half_transparent_mask_) {
        XRenderFreePicture(canvas_->xdisplay(), half_transparent_mask_);
        half_transparent_mask_ = 0;
    }
}

void
CompositeTestXRender::prepare_for_run(std::list<CompositeWindow*> &window_list)
{
    CompositeTest::prepare_for_run(window_list);

    if (options_["filter"].value == "bilinear")
        xrender_filter_ = FilterBilinear;
    else
        xrender_filter_ = FilterNearest;
}

void
CompositeTestXRender::draw(std::list<CompositeWindow *> &window_list)
{
    CompositeCanvasXRender *canvas_xrender =
        static_cast<CompositeCanvasXRender *>(canvas_);
    Display *xdpy = canvas_->xdisplay();

    Picture dest_picture = canvas_xrender->xrender_picture();
    int dest_width = canvas_->width();
    int dest_height = canvas_->height();

    /* Clear the screen with color matching GL {0.1, 0.1, 0.3, 1.0} */
    XRenderColor clear = {6554, 6554, 19661, 65535};

    XRenderFillRectangle(xdpy, PictOpSrc, dest_picture, &clear,
                         0, 0, dest_width, dest_height);

    /* Find out how many windows are visible and calculate the angular step */
    int visible_windows(num_visible_windows(window_list));
    float angular_step(2 * M_PI / visible_windows);

    /* Draw the windows in a circle using the calculated angular step */
    int i(0);
    for(std::list<CompositeWindow*>::iterator iter = window_list.begin();
        iter != window_list.end(); ++iter)
    {
        CompositeWindow *comp_win = *iter;
        Picture xrender_picture = static_cast<Picture>(comp_win->get_texture().i);
        if (xrender_picture) {
            int width = comp_win->width();
            int height = comp_win->height();

            /* Calculate the new position of the composited window center */
            int x0 = dest_width / 2 + cos(i * angular_step) * dest_width / 4;
            int y0 = dest_height / 2 + sin(i * angular_step) * dest_height / 4;

            /* Calculate the new position of the upper left corner */
            x0 -= dest_width / 4;
            y0 -= dest_height / 4;

            /* Scale the window to fit in dest_width/2 and dest_height/2 */
            XTransform xform = {{{0}}};
            xform.matrix[0][0] = XDoubleToFixed(2.0 * width / dest_width);
            xform.matrix[1][1] = XDoubleToFixed(2.0 * height / dest_height);
            xform.matrix[2][2] = XDoubleToFixed(1.0);

            XRenderSetPictureFilter(xdpy, xrender_picture,
                                    xrender_filter_, NULL, 0);
            XRenderSetPictureTransform(xdpy, xrender_picture, &xform);

            /* Draw the window Picture to the canvas Picture */
            XRenderComposite(xdpy, PictOpOver, xrender_picture,
                             half_transparent_mask_, dest_picture,
                             0, 0, 0, 0, x0, y0,
                             dest_width / 2, dest_height / 2);

            i++;
        }
    }
}


